import logging
import os
import shutil
import tempfile
from pathlib import Path
from typing import Any, Optional, Union

import mlflow
from mlflow.entities import Run as MLFlowRun
from mlflow.tracking.client import MlflowClient
from tango.common.aliases import PathOrStr
from tango.common.file_lock import AcquireReturnProxy, FileLock
from tango.common.params import Params
from tango.common.util import tango_cache_dir
from tango.step import Step
from tango.step_cache import CacheMetadata, StepCache
from tango.step_caches.local_step_cache import LocalStepCache
from tango.step_info import StepInfo

from tango_mlflow.util import (
    RunKind,
    get_mlflow_local_artifact_storage_path,
    get_mlflow_run_by_tango_step,
    get_mlflow_runs,
)

logger = logging.getLogger(__name__)


@StepCache.register("mlflow")
class MLFlowStepCache(LocalStepCache):
    def __init__(self, experiment_name: str) -> None:
        super().__init__(tango_cache_dir() / "mlflow_cache")
        self.experiment_name = experiment_name

    @property
    def mlflow_client(self) -> MlflowClient:
        return MlflowClient()

    def step_dir(self, step: Union[Step, StepInfo, str]) -> Path:
        mlflow_run = get_mlflow_run_by_tango_step(
            self.mlflow_client,
            self.experiment_name,
            tango_step=step,
        )
        if mlflow_run is None:
            return super().step_dir(step)

        mlflow_local_artifact_storage_path = get_mlflow_local_artifact_storage_path(mlflow_run)

        if mlflow_local_artifact_storage_path is None:
            return super().step_dir(step)

        return mlflow_local_artifact_storage_path

    def get_step_result_mlflow_run(self, step: Union[Step, StepInfo]) -> Optional[MLFlowRun]:
        return get_mlflow_run_by_tango_step(
            self.mlflow_client,
            self.experiment_name,
            tango_step=step,
            additional_filter_string="attributes.status = 'FINISHED'",
        )

    def create_step_result_artifact(
        self,
        step: Union[Step, StepInfo],
        objects_dir: Optional[PathOrStr] = None,
    ) -> None:
        mlflow_run = get_mlflow_run_by_tango_step(
            self.mlflow_client,
            self.experiment_name,
            tango_step=step,
        )
        if mlflow_run is None:
            raise RuntimeError(f"No MLflow run found for step: {step.unique_id}")

        if objects_dir is not None:
            self.mlflow_client.log_artifacts(mlflow_run.info.run_id, objects_dir)

    def _acquire_step_lock_file(
        self,
        step: Union[Step, StepInfo],
        read_only_ok: bool = False,
    ) -> AcquireReturnProxy:
        return FileLock(self.step_dir(step).with_suffix(".lock"), read_only_ok=read_only_ok).acquire_with_updates(
            desc=f"acquiring step cache lock for '{step.unique_id}'"
        )

    def __contains__(self, step: Any) -> bool:
        if not isinstance(step, (Step, StepInfo)):
            return False

        cacheable = step.cache_results if isinstance(step, Step) else step.cacheable
        if not cacheable:
            return False

        key = step.unique_id
        if key in self.strong_cache:
            return True
        if key in self.weak_cache:
            return True

        mlflow_run = self.get_step_result_mlflow_run(step)
        return mlflow_run is not None

    def __getitem__(self, step: Union[Step, StepInfo]) -> Any:
        key = step.unique_id

        # Try getting the result from our in-memory caches first.
        result = self._get_from_cache(key)
        if result is not None:
            return result

        def load_and_return() -> Any:
            metadata = CacheMetadata.from_params(Params.from_file(self._metadata_path(step)))
            result = metadata.format.read(self.step_dir(step))
            self._add_to_cache(key, result)
            return result

        # Next check our local on-disk cache
        with self._acquire_step_lock_file(step, read_only_ok=True):
            if self.step_dir(step).is_dir():
                return load_and_return()

        # Finally, check MLflow for the corresponding artifact.
        with self._acquire_step_lock_file(step):
            # Make sure the step wasn't cached since the last time we checked (above).
            if self.step_dir(step).is_dir():
                return load_and_return()

            mlflow_run = self.get_step_result_mlflow_run(step)
            if mlflow_run is None:
                raise KeyError(step)

            with tempfile.TemporaryDirectory(dir=self.dir, prefix=key) as temp_dir:
                mlflow.artifacts.download_artifacts(
                    run_id=mlflow_run.info.run_id,
                    artifact_path="",
                    dst_path=temp_dir,
                )
                os.replace(temp_dir, self.step_dir(step))

            return load_and_return()

    def __setitem__(self, step: Step, value: Any) -> None:
        if not step.cache_results:
            logger.warning("Tried to cache step %s despite being marked as uncacheable.", step.name)
            return

        with self._acquire_step_lock_file(step):
            with tempfile.TemporaryDirectory(dir=self.dir, prefix=step.unique_id) as _temp_dir:
                temp_dir = Path(_temp_dir)
                step.format.write(value, temp_dir)
                metadata = CacheMetadata(step=step.unique_id, format=step.format)
                metadata.to_params().to_file(temp_dir / self.METADATA_FILE_NAME)
                self.create_step_result_artifact(step, temp_dir)
                if self.step_dir(step).is_dir():
                    shutil.rmtree(self.step_dir(step), ignore_errors=True)
                os.replace(temp_dir, self.step_dir(step))

        self._add_to_cache(step.unique_id, value)

    def __len__(self) -> int:
        mlflow_runs = get_mlflow_runs(
            self.mlflow_client,
            self.experiment_name,
            run_kind=RunKind.STEP,
            additional_filter_string="attributes.status = 'FINISHED'",
        )
        return len(mlflow_runs)



from contextlib import suppress
from typing import Any, Dict, List, Optional

import mlflow
from mlflow.entities import Metric
from mlflow.entities import Run as MlflowRun
from mlflow.tracking.context import registry as context_registry
from mlflow.utils.mlflow_tags import MLFLOW_RUN_NAME
from tango.common.exceptions import IntegrationMissingError

from tango_mlflow.util import RunKind, flatten_dict, get_mlflow_run_by_tango_step, get_timestamp
from tango_mlflow.workspace import MLFlowWorkspace

with suppress(ModuleNotFoundError, IntegrationMissingError):
    import torch
    from tango.integrations.torch.train_callback import TrainCallback
    from tango.integrations.torch.util import peak_gpu_memory

    @TrainCallback.register("mlflow::log")
    class MLFlowTrainCallback(TrainCallback):
        def __init__(
            self,
            *args: Any,
            experiment_name: Optional[str] = None,
            tags: Optional[Dict[str, str]] = None,
            tracking_uri: Optional[str] = None,
            mlflow_config: Optional[Dict[str, Any]] = None,
            **kwargs: Any,
        ) -> None:
            if torch is None:
                raise IntegrationMissingError("torch")

            super().__init__(*args, **kwargs)

            if isinstance(self.workspace, MLFlowWorkspace):
                experiment_name = experiment_name or self.workspace.experiment_name
                tracking_uri = tracking_uri or self.workspace.mlflow_tracking_uri

            self.experiment_name = experiment_name
            self.tags = tags or {}
            self.tracking_uri = tracking_uri

            self.mlflow_config = self.train_config.as_dict()
            self.mlflow_config.pop("worker_id")
            if mlflow_config is not None:
                self.mlflow_config.update(mlflow_config)

            self.mlflow_run: Optional[MlflowRun] = None

        @property
        def mlflow_client(self) -> mlflow.tracking.MlflowClient:
            if isinstance(self.workspace, MLFlowWorkspace):
                return self.workspace.mlflow_client
            return mlflow.tracking.MlflowClient(tracking_uri=self.tracking_uri)

        def ensure_mlflow_run(self) -> MlflowRun:
            if self.mlflow_run is None:
                raise RuntimeError("MLFlow run not initialized")
            return self.mlflow_run

        def state_dict(self) -> Dict[str, Any]:
            return {}

        def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
            pass

        def pre_train_loop(self) -> None:
            if isinstance(self.workspace, MLFlowWorkspace):
                # Use existing MLFlow run created by the MLFlowWorkspace
                self.mlflow_run = get_mlflow_run_by_tango_step(
                    self.mlflow_client,
                    experiment=self.experiment_name,
                    tango_step=self.step_id,
                    additional_filter_string="attributes.status = 'RUNNING'",
                )
                if self.mlflow_run is None:
                    raise RuntimeError(f"Could not find a running MLFlow run for step {self.step_id}")
            else:
                # Create a new MLFlow run and log the config
                self.mlflow_run = self.mlflow_client.create_run(
                    experiment_id=mlflow.get_experiment_by_name(self.experiment_name).experiment_id,
                    tags=context_registry.resolve_tags(
                        {
                            "job_type": RunKind.TRAIN_METRICS,
                            "step_name": self.step_name,
                            "step_id": self.step_id,
                            MLFLOW_RUN_NAME: self.step_name,
                        }.update(self.tags)
                    ),
                )
                for key, value in flatten_dict(self.mlflow_config).items():
                    self.mlflow_client.log_param(self.mlflow_run.info.run_id, key, value)

            if torch.cuda.is_available():
                torch.cuda.reset_peak_memory_stats()
            peak_gpu_mbs = peak_gpu_memory()
            if self.is_local_main_process:
                timestamp = self.mlflow_run.info.start_time
                metrics = [
                    Metric(key=f"sys.worker{rank}_peak_gpu_mem", value=mbs, timestamp=timestamp, step=0)
                    for rank, mbs in peak_gpu_mbs.items()
                ]
                metrics.append(Metric(key="epoch", value=0, timestamp=timestamp, step=0))
                self.mlflow_client.log_batch(self.mlflow_run.info.run_id, metrics=metrics)

        def post_train_loop(self, step: int, epoch: int) -> None:
            if isinstance(self.workspace, MLFlowWorkspace):
                # We don't need to do anything here, as the MLFlow run will be closed by the MLFlowWorkspace
                return
            if self.is_local_main_process:
                mlflow_run = self.ensure_mlflow_run()
                self.mlflow_client.set_terminated(mlflow_run.info.run_id)

        def log_batch(
            self,
            step: int,
            epoch: int,
            batch_loss: float,
            batch_outputs: List[Dict[str, Any]],
        ) -> None:
            step = step + 1
            timestamp = get_timestamp()
            peak_gpu_mbs = peak_gpu_memory()
            if self.is_local_main_process:
                mlflow_run = self.ensure_mlflow_run()
                metrics = [
                    Metric(
                        key="train.loss",
                        value=batch_loss,
                        timestamp=timestamp,
                        step=step,
                    ),
                    Metric(
                        key="train.lr",
                        value=self.training_engine.optimizer.param_groups[0]["lr"],
                        timestamp=timestamp,
                        step=step,
                    ),
                    Metric(
                        key="epoch",
                        value=epoch,
                        timestamp=0,
                        step=step,
                    ),
                ] + [
                    Metric(key=f"sys.worker{rank}_peak_gpu_mem", value=mbs, timestamp=timestamp, step=step)
                    for rank, mbs in peak_gpu_mbs.items()
                ]
                self.mlflow_client.log_batch(mlflow_run.info.run_id, metrics=metrics)

        def post_val_loop(
            self,
            step: int,
            epoch: int,
            val_metric: float,
            best_val_metric: float,
        ) -> None:
            step = step + 1
            timestamp = get_timestamp()
            if self.is_local_main_process:
                mlflow_run = self.ensure_mlflow_run()
                metrics = [
                    Metric(
                        key="val.loss",
                        value=val_metric,
                        timestamp=timestamp,
                        step=step,
                    ),
                    Metric(
                        key=f"val.best_{self.train_config.val_metric_name}",
                        value=best_val_metric,
                        timestamp=timestamp,
                        step=step,
                    ),
                    Metric(
                        key="epoch",
                        value=epoch,
                        timestamp=timestamp,
                        step=step,
                    ),
                ]
                self.mlflow_client.log_batch(mlflow_run.info.run_id, metrics=metrics)



import typing
from typing import Any, Dict, Optional, Protocol, TypeVar

from mlflow.entities import Run as MLflowRun
from mlflow.tracking import MlflowClient
from tango.step import Step

T = TypeVar("T")


@typing.runtime_checkable
class MLflowSummaryStep(Protocol):
    MLFLOW_SUMMARY: bool


class MLflowLogger:
    def __init__(self, mlflow_run: MLflowRun):
        self._mlflow_run = mlflow_run

    @property
    def mlflow_run(self) -> MLflowRun:
        return self._mlflow_run

    @property
    def mlflow_client(self) -> MlflowClient:
        return MlflowClient()

    def log_metric(
        self,
        key: str,
        value: float,
        timestamp: Optional[int] = None,
        step: Optional[int] = None,
    ) -> None:
        self.mlflow_client.log_metric(
            run_id=self.mlflow_run.info.run_id,
            key=key,
            value=value,
            timestamp=timestamp,
            step=step,
        )

    def log_metrics(self, metrics: Dict[str, float]) -> None:
        for key, value in metrics.items():
            self.log_metric(key, value)


class MLflowStep(Step[T]):
    MLFLOW_SUMMARY: bool = False

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._mlflow_run: Optional[MLflowRun] = None
        self._mlflow_logger: Optional[MLflowLogger] = None

    def setup_mlflow(self, mlflow_run: MLflowRun) -> None:
        if self._mlflow_run is not None:
            raise RuntimeError("MLflow run already set")
        self._mlflow_run = mlflow_run
        self._mlflow_logger = MLflowLogger(self._mlflow_run)

    @property
    def mlflow_run(self) -> MLflowRun:
        if self._mlflow_run is None:
            raise RuntimeError("MLflow run not set")
        return self._mlflow_run

    @property
    def mlflow_logger(self) -> MLflowLogger:
        if self._mlflow_logger is None:
            raise RuntimeError("MLflow logger not set")
        return self._mlflow_logger



from __future__ import annotations

import argparse

from tango_mlflow import __version__
from tango_mlflow.commands import tune  # noqa: F401
from tango_mlflow.commands.subcommand import Subcommand


def create_subcommand(prog: str | None = None) -> Subcommand:
    parser = argparse.ArgumentParser(usage="%(prog)s", prog=prog)
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s " + __version__,
    )
    return Subcommand(parser)


def main(prog: str | None = None) -> None:
    app = create_subcommand(prog)
    app()



import argparse
import copy
import dataclasses
import json
import os
import tempfile
from functools import partial
from logging import getLogger
from typing import Any, Dict, List, Optional, cast

import mlflow
from mlflow.utils.mlflow_tags import MLFLOW_PARENT_RUN_ID, MLFLOW_RUN_NAME
from tango.__main__ import _run as tango_run
from tango.common import FromParams, Params, Registrable
from tango.common.logging import initialize_logging
from tango.common.util import import_extra_module
from tango.settings import TangoGlobalSettings
from tango.workspace import Workspace

from tango_mlflow.commands.subcommand import Subcommand
from tango_mlflow.util import (
    RunKind,
    build_filter_string,
    generate_unique_run_name,
    get_mlflow_run_by_tango_run,
    is_all_child_run_finished,
)
from tango_mlflow.workspace import MLFlowWorkspace

logger = getLogger(__name__)


@Subcommand.register("tune")
class TuneCommand(Subcommand):
    """Tune hyper-parameters using Optuna."""

    def setup(self) -> None:
        self.parser.add_argument(
            "experiment",
            type=str,
        )
        self.parser.add_argument(
            "hparam_path",
            type=str,
            help="path to hyperparameter file",
        )
        self.parser.add_argument(
            "--name",
            type=str,
            help="name of optuna study",
        )
        self.parser.add_argument(
            "--tango-settings",
            type=str,
            help="path to tango settings file",
        )
        self.parser.add_argument(
            "--metric",
            type=str,
            help="metric you want to optimize",
        )
        self.parser.add_argument(
            "--n-trials",
            type=int,
            help="The number of trials. If this argument is not given, as many trials run as possible.",
        )
        self.parser.add_argument(
            "--direction",
            type=str,
            choices=("minimize", "maximize"),
            help="Set direction of optimization to a new study. Set 'minimize' "
            "for minimization and 'maximize' for maximization.",
        )
        self.parser.add_argument(
            "--timeout",
            type=float,
            help="Stop study after the given number of second(s). If this argument"
            " is not given, as many trials run as possible.",
        )
        self.parser.add_argument(
            "--resume",
            action="store_true",
            help="Resume optimization if a study with the same name already exists.",
        )
        self.parser.add_argument(
            "--optuna-settings",
            type=str,
            default="optuna.yml",
            help="path to optuna settings file",
        )

    def run(self, args: argparse.Namespace) -> None:
        import optuna
        from optuna.pruners import BasePruner
        from optuna.samplers import BaseSampler

        class OptunaPruner(BasePruner, Registrable):  # type: ignore[misc]
            ...

        OptunaPruner.register("median")(optuna.pruners.MedianPruner)
        OptunaPruner.register("nop")(optuna.pruners.NopPruner)
        OptunaPruner.register("percentile")(optuna.pruners.PercentilePruner)
        OptunaPruner.register("patient")(optuna.pruners.PatientPruner)
        OptunaPruner.register("successive_halving")(optuna.pruners.SuccessiveHalvingPruner)
        OptunaPruner.register("threshold")(optuna.pruners.ThresholdPruner)
        OptunaPruner.register("hyperband")(optuna.pruners.HyperbandPruner)

        class OptunaSampler(BaseSampler, Registrable):  # type: ignore[misc]
            ...

        OptunaSampler.register("grid")(optuna.samplers.GridSampler)
        OptunaSampler.register("random")(optuna.samplers.RandomSampler)
        OptunaSampler.register("tpe")(optuna.samplers.TPESampler)
        OptunaSampler.register("cmaes")(optuna.samplers.CmaEsSampler)
        OptunaSampler.register("partial_fixed")(optuna.samplers.PartialFixedSampler)
        OptunaSampler.register("nsga2")(optuna.samplers.NSGAIISampler)
        OptunaSampler.register("motpe")(optuna.samplers.MOTPESampler)
        OptunaSampler.register("qmcsampler")(optuna.samplers.QMCSampler)

        @dataclasses.dataclass
        class OptunaSettings(FromParams):
            """Settings for Optuna."""

            metric: Optional[str] = None
            direction: str = "minimize"
            sampler: Optional[OptunaSampler] = None
            pruner: Optional[OptunaPruner] = None

            _original_params: Optional[Params] = None

            @classmethod
            def from_args(cls, args: argparse.Namespace) -> "OptunaSettings":
                if args.optuna_settings is not None and os.path.exists(args.optuna_settings):
                    logger.info(f"Loading Optuna settings from {args.optuna_settings}")
                    params = Params.from_file(args.optuna_settings)
                else:
                    params = Params({})
                settings = cls.from_params(copy.deepcopy(params), _original_params=params)
                if args.metric is not None:
                    settings.metric = args.metric
                if args.direction is not None:
                    settings.direction = args.direction
                return settings

            def _to_params(self) -> Dict[str, Any]:
                if self._original_params is not None:
                    return cast(Dict[str, Any], self._original_params.as_dict())
                return super()._to_params()

        def get_mlflow_run_by_optuna_study(
            mlflow_client: mlflow.tracking.MlflowClient,
            experiment_name: str,
            study_name: str,
        ) -> Optional[mlflow.entities.Run]:
            experiment = mlflow.get_experiment_by_name(experiment_name)
            if experiment is None:
                raise ValueError(f"Experiment {experiment_name} does not exist.")
            mlflow_runs = mlflow_client.search_runs(
                experiment_ids=[experiment.experiment_id],
                filter_string=build_filter_string(
                    run_kind=RunKind.OPTUNA_STUDY,
                    additional_filter_string=f"tags.{MLFLOW_RUN_NAME} = '{study_name}'",
                ),
            )
            if not mlflow_runs:
                return None
            if len(mlflow_runs) > 1:
                raise ValueError(f"Found more than one run name '{study_name}' in MLflow experiment")
            return mlflow_runs[0]

        def _objective(
            trial: optuna.trial.Trial,
            metric: str,
            experiment: str,
            hparam_path: str,
            tango_settings: TangoGlobalSettings,
        ) -> float:
            tango_settings = copy.deepcopy(tango_settings)

            ext_var: List[str] = []
            with open(hparam_path) as jsonfile:
                for key, params in json.load(jsonfile).items():
                    value_type = params.pop("type")
                    suggest = getattr(trial, f"suggest_{value_type}")
                    value = suggest(key, **params)
                    ext_var.append(f"{key}={value}")

            run_name = tango_run(
                settings=copy.deepcopy(tango_settings),
                experiment=experiment,
                name=f"{trial.study.study_name}-{trial.number}",
                ext_var=ext_var,
            )

            workspace = Workspace.from_params(Params(tango_settings.workspace or {}))
            assert isinstance(workspace, MLFlowWorkspace)

            mlflow_run = get_mlflow_run_by_tango_run(
                workspace.mlflow_client,
                workspace.experiment_name,
                tango_run=run_name,
            )
            if mlflow_run is None:
                raise RuntimeError(f"Could not find MLFlow run for tango run {run_name}")

            if mlflow_run.info.status == "RUNNING":
                status = (
                    "FINISHED"
                    if is_all_child_run_finished(
                        workspace.mlflow_client,
                        workspace.experiment_name,
                        mlflow_run=mlflow_run,
                    )
                    else "FAILED"
                )
                workspace.mlflow_client.set_terminated(mlflow_run.info.run_id, status)

            metric_history = workspace.mlflow_client.get_metric_history(mlflow_run.info.run_id, metric)
            if not metric_history:
                raise RuntimeError(f"No metric history found for {args.metric}.")

            latest_metric = metric_history[-1]

            workspace.mlflow_client.log_metric(
                mlflow_run.data.tags[MLFLOW_PARENT_RUN_ID],
                metric,
                latest_metric.value,
                step=trial.number,
            )

            return float(latest_metric.value)

        tango_settings = (
            TangoGlobalSettings.from_file(args.tango_settings) if args.tango_settings else TangoGlobalSettings.default()
        )

        for package_name in tango_settings.include_package or ():
            import_extra_module(package_name)

        initialize_logging(
            log_level=tango_settings.log_level,
            enable_cli_logs=True,
            file_friendly_logging=tango_settings.file_friendly_logging,
        )

        workspace = Workspace.from_params(tango_settings.workspace or {})
        if not isinstance(workspace, MLFlowWorkspace):
            raise ValueError("Tango workspace type must be mlflow.")

        optuna_settings = OptunaSettings.from_args(args)
        if optuna_settings.metric is None:
            raise ValueError("Optuna metric must be specified.")
        study_name = args.name or generate_unique_run_name(workspace.mlflow_client, workspace.experiment_name)

        with tempfile.TemporaryDirectory() as workdir:
            optuna_storage_path = os.path.join(workdir, "optuna.db")

            # Load previous study if exists
            mlflow_run = get_mlflow_run_by_optuna_study(
                workspace.mlflow_client,
                workspace.experiment_name,
                study_name=study_name,
            )
            if args.resume and mlflow_run is not None:
                logger.info(f"Resuming run {args.name}.")
                mlflow.artifacts.download_artifacts(
                    run_id=mlflow_run.info.run_id,
                    artifact_path="optuna.db",
                    dst_path=workdir,
                )
                assert os.path.exists(optuna_storage_path)
            elif not args.resume and mlflow_run is not None:
                raise ValueError(f"Run {args.name} already exists. If you want to resume, use --resume option.")
            elif args.resume and mlflow_run is None:
                raise ValueError(f"Resume requested, but run {args.name} does not exist.")

            study = optuna.create_study(
                study_name=study_name,
                direction=optuna_settings.direction,
                sampler=optuna_settings.sampler,
                pruner=optuna_settings.pruner,
                storage=f"sqlite:///{optuna_storage_path}",
                load_if_exists=args.resume,
            )

            with mlflow.start_run(
                run_id=mlflow_run.info.run_id if mlflow_run is not None else None,
                run_name=study_name if mlflow_run is None else None,
                tags={"job_type": RunKind.OPTUNA_STUDY.value},
            ) as active_run:
                workspace._mlflow_tags[MLFLOW_PARENT_RUN_ID] = active_run.info.run_id
                tango_settings.workspace = workspace.to_params().as_dict()

                mlflow.log_params(optuna_settings.to_params().as_flat_dict())

                objective = partial(
                    _objective,
                    metric=optuna_settings.metric,
                    experiment=args.experiment,
                    hparam_path=args.hparam_path,
                    tango_settings=tango_settings,
                )

                mlflow.log_artifact(args.experiment)
                mlflow.log_artifact(args.hparam_path)

                try:
                    study.optimize(
                        objective,
                        n_trials=args.n_trials,
                        timeout=args.timeout,
                    )
                    best_result = {
                        "trial": study.best_trial.number,
                        "params": study.best_trial.params,
                        "metric": study.best_trial.value,
                    }
                    best_result_filename = os.path.join(workdir, "best.json")
                    with open(best_result_filename, "w") as jsonfile:
                        json.dump(best_result, jsonfile)
                    mlflow.log_artifact(best_result_filename)
                finally:
                    mlflow.log_artifact(optuna_storage_path)



import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Union
from urllib.parse import urlencode, urlparse

import mlflow
import petname
from mlflow.entities import Experiment as MLFlowExperiment
from mlflow.entities import Run as MLFlowRun
from mlflow.tracking import artifact_utils
from mlflow.tracking.client import MlflowClient
from mlflow.tracking.context import registry as context_registry
from mlflow.utils.mlflow_tags import MLFLOW_PARENT_RUN_ID, MLFLOW_RUN_NAME, MLFLOW_RUN_NOTE
from tango.step import Step
from tango.step_info import StepInfo
from tango.workspace import Run as TangoRun


class RunKind(Enum):
    STEP = "step"
    TANGO_RUN = "tango_run"
    OPTUNA_STUDY = "optuna_study"
    TRAIN_METRICS = "train_metrics"


def flatten_dict(d: Dict[str, Any]) -> Dict[str, Any]:
    """
    Flatten a nested dictionary into a single level dictionary.
    """
    result = {}
    for key, value in d.items():
        if isinstance(value, (list, tuple)):
            value = {str(i): v for i, v in enumerate(value)}
        if isinstance(value, dict):
            result.update({f"{key}.{k}": v for k, v in flatten_dict(value).items()})
        else:
            result[key] = value
    return result


def get_timestamp() -> int:
    """timestamp as an integer (milliseconds since the Unix epoch)."""
    return int(datetime.datetime.now().timestamp() * 1000)


def is_mlflow_using_local_artifact_storage(
    mlflow_run: Union[str, MLFlowRun],
) -> bool:
    mlflow_run_id = mlflow_run.info.run_id if isinstance(mlflow_run, MLFlowRun) else mlflow_run
    return str(artifact_utils.get_artifact_uri(run_id=mlflow_run_id)).startswith("file")


def get_mlflow_local_artifact_storage_path(
    mlflow_run: Union[str, MLFlowRun],
) -> Optional[Path]:
    mlflow_run_id = mlflow_run.info.run_id if isinstance(mlflow_run, MLFlowRun) else mlflow_run
    mlflow_artifact_uri = urlparse(artifact_utils.get_artifact_uri(run_id=mlflow_run_id))
    if mlflow_artifact_uri.scheme == "file":
        return Path(mlflow_artifact_uri.path)
    return None


def build_filter_string(
    run_kind: Optional[RunKind] = None,
    tango_step: Optional[Union[str, Step, StepInfo]] = None,
    tango_run: Optional[Union[str, TangoRun]] = None,
    additional_filter_string: Optional[str] = None,
) -> str:
    conditions: List[str] = []
    if run_kind is not None:
        conditions.append(f"tags.job_type = '{run_kind.value}'")
    if tango_step is not None:
        tango_step_id = tango_step if isinstance(tango_step, str) else tango_step.unique_id
        if run_kind == RunKind.TANGO_RUN:
            conditions.append(f"params._step_ids.\"{tango_step_id}\" = 'True'")
        else:
            conditions.append(f"tags.step_id = '{tango_step_id}'")
    if tango_run is not None:
        tango_run_name = tango_run.name if isinstance(tango_run, TangoRun) else tango_run
        conditions.append(f"tags.{MLFLOW_RUN_NAME} = '{tango_run_name}'")
    if additional_filter_string is not None:
        conditions.append(additional_filter_string)
    return " AND ".join(conditions)


def get_mlflow_run_by_tango_step(
    mlflow_client: MlflowClient,
    experiment: Union[str, MLFlowExperiment],
    tango_step: Union[str, Step, StepInfo],
    additional_filter_string: Optional[str] = None,
) -> Optional[MLFlowRun]:
    if isinstance(experiment, str):
        experiment = mlflow_client.get_experiment_by_name(experiment)

    runs = mlflow_client.search_runs(
        experiment_ids=[experiment.experiment_id],
        filter_string=build_filter_string(
            run_kind=RunKind.STEP,
            tango_step=tango_step,
            additional_filter_string=additional_filter_string,
        ),
    )

    if not runs:
        return None

    return runs[0]


def get_mlflow_run_by_tango_run(
    mlflow_client: MlflowClient,
    experiment: Union[str, MLFlowExperiment],
    tango_run: Union[str, TangoRun],
    additional_filter_string: Optional[str] = None,
) -> Optional[MLFlowRun]:
    if isinstance(experiment, str):
        experiment = mlflow.get_experiment_by_name(experiment)

    tango_run_name = tango_run.name if isinstance(tango_run, TangoRun) else tango_run

    mlflow_runs = mlflow_client.search_runs(
        experiment_ids=[experiment.experiment_id],
        filter_string=build_filter_string(
            run_kind=RunKind.TANGO_RUN,
            tango_run=tango_run_name,
            additional_filter_string=additional_filter_string,
        ),
    )

    if not mlflow_runs:
        return None

    if len(mlflow_runs) > 1:
        raise ValueError(f"Found more than one run name '{tango_run_name}' in MLflow experiment")

    return mlflow_runs[0]


def get_mlflow_runs(
    mlflow_client: MlflowClient,
    experiment: Union[str, MLFlowExperiment],
    run_kind: Optional[RunKind] = None,
    tang_run: Optional[Union[str, TangoRun]] = None,
    tango_step: Optional[Union[str, Step, StepInfo]] = None,
    additional_filter_string: Optional[str] = None,
) -> List[MLFlowRun]:
    if isinstance(experiment, str):
        experiment = mlflow.get_experiment_by_name(experiment)

    mlflow_runs: List[MLFlowRun] = []
    page_token: Optional[str] = None
    while True:
        matching_runs = mlflow_client.search_runs(
            experiment_ids=[experiment.experiment_id],
            filter_string=build_filter_string(
                run_kind=run_kind,
                tango_step=tango_step,
                tango_run=tang_run,
                additional_filter_string=additional_filter_string,
            ),
            page_token=page_token,
        )
        mlflow_runs.extend(matching_runs)
        if matching_runs.token is None:
            break
        page_token = matching_runs.token

    return mlflow_runs


def generate_unique_run_name(
    mlflow_client: MlflowClient,
    experiment: Union[str, MLFlowExperiment],
) -> str:
    if isinstance(experiment, str):
        experiment = mlflow.get_experiment_by_name(experiment)
    name: Optional[str] = None
    while name is None or get_mlflow_run_by_tango_run(mlflow_client, experiment, name) is not None:
        name = petname.generate()
    return name


def add_mlflow_run_of_tango_run(
    mlflow_client: MlflowClient,
    experiment: Union[str, MLFlowExperiment],
    steps: Set[Step],
    run_name: Optional[str] = None,
    mlflow_tags: Optional[Dict[str, str]] = None,
) -> MLFlowRun:
    if isinstance(experiment, str):
        experiment = mlflow.get_experiment_by_name(experiment)

    # generate a unique run name
    if run_name is None:
        run_name = generate_unique_run_name(mlflow_client, experiment)

    # build description of the run
    description = "# Tango run"
    cacheable_steps = {step for step in steps if step.cache_results}
    if cacheable_steps:
        description += "\nCacheable steps:\n"
        for step in sorted(cacheable_steps, key=lambda step: step.name):
            description += f"- {step.name}"
            dependencies = step.dependencies
            if dependencies:
                description += ", depends on: " + ", ".join(
                    sorted(
                        [f"'{dep.name}'" for dep in dependencies],
                    )
                )
            search_query_key = "searchInput" if mlflow.__version__ < "2.0.0" else "searchFilter"
            query = {search_query_key: f"tags.step_id = '{step.unique_id}'"}
            url = f"/#/experiments/{experiment.experiment_id}/s?{urlencode(query)}"
            description += f"\n  - [{step.unique_id}]({url})\n"

    mlflow_run = mlflow_client.create_run(
        experiment_id=experiment.experiment_id,
        tags=context_registry.resolve_tags(
            {
                "job_type": RunKind.TANGO_RUN.value,
                MLFLOW_RUN_NAME: run_name,
                MLFLOW_RUN_NOTE: description,
                **(mlflow_tags or {}),
            }
        ),
        run_name=run_name,
    )

    step_ids: Dict[str, bool] = {}
    step_name_to_info: Dict[str, Dict[str, Any]] = {}
    for step in steps:
        step_info = StepInfo.new_from_step(step)
        step_name_to_info[step.name] = {k: v for k, v in step_info.to_json_dict().items() if v is not None}

        # log step config
        step_config = flatten_dict({step.name: step.config})
        for key, value in step_config.items():
            mlflow_client.log_param(mlflow_run.info.run_id, key, value)

        # log dependent steps
        step_ids[step.unique_id] = True
        mlflow_client.log_param(mlflow_run.info.run_id, f"_step_ids.{step.unique_id}", True)

    # log entire step info
    mlflow_client.log_dict(
        run_id=mlflow_run.info.run_id,
        dictionary={"steps": step_name_to_info, "_step_ids": step_ids},
        artifact_file="steps.json",
    )

    return mlflow_run


def add_mlflow_run_of_tango_step(
    mlflow_client: MlflowClient,
    experiment: Union[str, MLFlowExperiment],
    tango_run: Union[str, TangoRun],
    step_info: StepInfo,
) -> MLFlowRun:
    if isinstance(experiment, str):
        experiment = mlflow.get_experiment_by_name(experiment)

    parent_mlflow_run = get_mlflow_run_by_tango_run(mlflow_client, experiment, tango_run)
    if parent_mlflow_run is None:
        raise ValueError(f"Could not find MLflow run for Tango run '{tango_run}'")

    description = "\n".join(
        (
            f"### Tango step ***{step_info.step_name}***",
            f"- type: {step_info.step_class_name}",
            f"- ID: {step_info.unique_id}",
        )
    )

    mlflow_run = mlflow_client.create_run(
        experiment_id=experiment.experiment_id,
        start_time=None if step_info.start_time is None else round(1000 * step_info.start_time.timestamp()),
        tags=context_registry.resolve_tags(
            {
                "job_type": RunKind.STEP.value,
                "step_name": step_info.step_name,
                "step_id": step_info.unique_id,
                MLFLOW_RUN_NAME: step_info.step_name,
                MLFLOW_PARENT_RUN_ID: parent_mlflow_run.info.run_id,
                MLFLOW_RUN_NOTE: description,
            }
        ),
        run_name=step_info.step_name,
    )

    for key, value in flatten_dict(step_info.config or {}).items():
        mlflow_client.log_param(mlflow_run.info.run_id, key, value)
    mlflow_client.log_dict(
        run_id=mlflow_run.info.run_id,
        dictionary=step_info.to_json_dict(),
        artifact_file="step_info.json",
    )

    return mlflow_run


def terminate_mlflow_run_of_tango_step(
    mlflow_client: MlflowClient,
    experiment: Union[str, MLFlowExperiment],
    status: str,
    step_info: StepInfo,
) -> None:
    if isinstance(experiment, str):
        experiment = mlflow.get_experiment_by_name(experiment)

    mlflow_run = get_mlflow_run_by_tango_step(mlflow_client, experiment, step_info)
    if mlflow_run is None:
        raise ValueError(f"Could not find MLFlow run for step {step_info.unique_id}")

    mlflow_client.log_dict(
        run_id=mlflow_run.info.run_id,
        dictionary=step_info.to_json_dict(),
        artifact_file="step_info.json",
    )
    mlflow_client.set_terminated(
        run_id=mlflow_run.info.run_id,
        status=status,
        end_time=None if step_info.end_time is None else round(1000 * step_info.end_time.timestamp()),
    )


def is_all_child_run_finished(
    mlflow_client: MlflowClient,
    experiment: Union[str, MLFlowExperiment],
    mlflow_run: Union[str, MLFlowRun],
) -> bool:
    run_id = mlflow_run.info.run_id if isinstance(mlflow_run, MLFlowRun) else mlflow_run

    for child_run in get_mlflow_runs(
        mlflow_client,
        experiment,
        additional_filter_string=f"tags.{MLFLOW_PARENT_RUN_ID} = '{run_id}'",
    ):
        if child_run.info.status != "FINISHED":
            return False
    return True



import atexit
import json
import logging
import tempfile
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import Any, ContextManager, Dict, Iterable, Iterator, Optional, TypeVar, Union, cast
from urllib.parse import ParseResult, parse_qs, quote

import mlflow
import pytz  # type: ignore
from mlflow.entities import Run as MLFlowRun
from mlflow.tracking.client import MlflowClient
from mlflow.utils.mlflow_tags import MLFLOW_RUN_NAME
from tango.common.exceptions import StepStateError
from tango.common.file_lock import FileLock
from tango.common.logging import file_handler
from tango.common.util import exception_to_string, tango_cache_dir, utc_now_datetime
from tango.step import Step
from tango.step_cache import StepCache
from tango.step_info import StepInfo, StepState
from tango.workspace import Run, Workspace

from tango_mlflow.step import MLflowStep, MLflowSummaryStep
from tango_mlflow.step_cache import MLFlowStepCache
from tango_mlflow.util import (
    RunKind,
    add_mlflow_run_of_tango_run,
    add_mlflow_run_of_tango_step,
    flatten_dict,
    get_mlflow_run_by_tango_run,
    get_mlflow_run_by_tango_step,
    get_mlflow_runs,
    terminate_mlflow_run_of_tango_step,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")


@Workspace.register("mlflow")
class MLFlowWorkspace(Workspace):
    def __init__(
        self,
        experiment_name: str,
        tags: Optional[Dict[str, str]] = None,
        tracking_uri: Optional[str] = None,
    ) -> None:
        mlflow.set_experiment(experiment_name)
        super().__init__()  # type: ignore[no-untyped-call]
        self.experiment_name = experiment_name
        self.cache = MLFlowStepCache(experiment_name=self.experiment_name)
        self.steps_dir = tango_cache_dir() / "mlflow_workspace"
        self.locks: Dict[Step, FileLock] = {}
        self._running_step_info: Dict[str, StepInfo] = {}
        self._step_id_to_run_name: Dict[str, str] = {}
        self._mlflow_tags = dict(sorted(tags.items())) if tags else {}
        self._mlflow_tracking_uri = tracking_uri
        self._run_status: Dict[str, str] = {}

        if self._mlflow_tracking_uri is not None:
            mlflow.set_tracking_uri(tracking_uri)

    def __getstate__(self) -> Dict[str, Any]:
        out = super().__getstate__()  # type: ignore[no-untyped-call]
        out["locks"] = {}
        return cast(Dict[str, Any], out)

    @property
    def url(self) -> str:
        url = f"mlflow://{self.experiment_name}"
        if self._mlflow_tags:
            url += f"?tags={quote(json.dumps(self._mlflow_tags))}"
        if self._mlflow_tracking_uri is not None:
            url += f"&tracking_uri={quote(self._mlflow_tracking_uri)}"
        return url

    @classmethod
    def from_parsed_url(cls, parsed_url: ParseResult) -> "MLFlowWorkspace":
        queries = parse_qs(parsed_url.query)
        experiment_name = parsed_url.netloc
        tags: Dict[str, Any] = {}
        for json_strng in queries.get("tags", []):
            subtags = json.loads(json_strng)
            if not isinstance(subtags, dict):
                raise ValueError(f"Invalid tags: {subtags}")
            tags.update(subtags)
        tracking_uri: Optional[str] = None
        for uri in queries.get("tracking_uri", []):
            tracking_uri = uri
        return cls(
            experiment_name=experiment_name,
            tags=tags,
            tracking_uri=tracking_uri,
        )

    @property
    def step_cache(self) -> StepCache:
        return self.cache

    @property
    def mlflow_client(self) -> MlflowClient:
        return MlflowClient()

    @property
    def mlflow_tracking_uri(self) -> str:
        return str(mlflow.get_tracking_uri()).rstrip("/")

    def _get_unique_id(self, step_or_unique_id: Union[Step, str]) -> str:
        if isinstance(step_or_unique_id, Step):
            return step_or_unique_id.unique_id
        return step_or_unique_id

    def step_dir(self, step_or_unique_id: Union[Step, str]) -> Path:
        unique_id = self._get_unique_id(step_or_unique_id)
        path = self.steps_dir / unique_id
        path.mkdir(parents=True, exist_ok=True)
        return path

    def work_dir(self, step: Step) -> Path:
        path = self.step_dir(step) / "work"
        path.mkdir(parents=True, exist_ok=True)
        return path

    def step_info(self, step_or_unique_id: Union[Step, str]) -> StepInfo:
        unique_id = self._get_unique_id(step_or_unique_id)
        if unique_id in self._running_step_info:
            return self._running_step_info[unique_id]
        step_info = self._get_updated_step_info(
            unique_id,
            step_name=step_or_unique_id.name if isinstance(step_or_unique_id, Step) else None,
        )
        if step_info is None:
            raise KeyError(step_or_unique_id)
        return step_info

    def step_starting(self, step: Step) -> None:
        lock_path = self.step_dir(step) / "lock"
        lock = FileLock(lock_path, read_only_ok=True)
        lock.acquire_with_updates(desc=f"acquiring lock for '{step.name}'")
        self.locks[step] = lock

        step_info = self._get_updated_step_info(step.unique_id) or StepInfo.new_from_step(step)
        if step_info.state not in {StepState.INCOMPLETE, StepState.FAILED, StepState.UNCACHEABLE}:
            raise StepStateError(
                step,
                step_info.state,
                context="If you are certain the step is not running somewhere else, delete the lock "
                f"file at {lock_path}.",
            )

        try:
            step_info.start_time = utc_now_datetime()
            step_info.end_time = None
            step_info.error = None
            step_info.result_location = None

            mlflow_run = add_mlflow_run_of_tango_step(
                self.mlflow_client,
                self.experiment_name,
                tango_run=self._step_id_to_run_name[step.unique_id],
                step_info=step_info,
            )

            if isinstance(step, MLflowStep):
                step.setup_mlflow(mlflow_run)

            logger.info(
                "Tracking '%s' step on MLflow: %s/#/experiments/%s/runs/%s",
                step.name,
                self.mlflow_tracking_uri,
                mlflow_run.info.experiment_id,
                mlflow_run.info.run_id,
            )

            self._running_step_info[step.unique_id] = step_info
        except:  # noqa: E722
            lock.release()
            del self.locks[step]
            raise

    def step_finished(self, step: Step, result: T) -> T:
        mlflow_run = get_mlflow_run_by_tango_step(self.mlflow_client, self.experiment_name, step)
        if mlflow_run is None:
            raise RuntimeError(
                f"{self.__class__.__name__}.step_finished() called outside of a MLflow run. "
                f"Did you forget to call {self.__class__.__name__}.step_starting() first?"
            )

        step_info = self._running_step_info.get(step.unique_id) or self._get_updated_step_info(step.unique_id)
        if step_info is None:
            raise KeyError(step.unique_id)

        try:
            if step.cache_results:
                self.step_cache[step] = result
                if hasattr(result, "__next__"):
                    assert isinstance(result, Iterator)
                    # Caching the iterator will consume it, so we write it to the
                    # cache and then read from the cache for the return value.
                    result = self.step_cache[step]
                step_info.result_location = mlflow_run.info.artifact_uri
            else:
                self.cache.create_step_result_artifact(step)

            step_info.end_time = utc_now_datetime()
            terminate_mlflow_run_of_tango_step(
                self.mlflow_client,
                self.experiment_name,
                status="FINISHED",
                step_info=step_info,
            )

            # Log the result of summary step to a parent mlflow run
            if isinstance(step, MLflowSummaryStep) and step.MLFLOW_SUMMARY:
                if not isinstance(result, dict):
                    raise ValueError(
                        f"Result value of Step {step.name} with MLFLOW_SUMMARY=True"
                        f"must be a dict, but got {type(result)}"
                    )

                mlflow_run_of_tang_run = get_mlflow_run_by_tango_run(
                    self.mlflow_client,
                    self.experiment_name,
                    tango_run=self._step_id_to_run_name[step.unique_id],
                )
                if mlflow_run_of_tang_run is None:
                    raise RuntimeError(
                        f"Could not find MLflow run for Tango run {self._step_id_to_run_name[step.unique_id]}"
                    )

                for key, value in flatten_dict({step.name: result}).items():
                    self.mlflow_client.log_metric(mlflow_run_of_tang_run.info.run_id, key, value)
        finally:
            self.locks[step].release()
            del self.locks[step]
            if step.unique_id in self._running_step_info:
                del self._running_step_info[step.unique_id]

        return result

    def step_failed(self, step: Step, e: BaseException) -> None:
        self._run_status[self._step_id_to_run_name[step.unique_id]] = "FAILED"
        mlflow_run = get_mlflow_run_by_tango_step(
            self.mlflow_client,
            self.experiment_name,
            tango_step=step,
        )
        if mlflow_run is None:
            raise RuntimeError(
                f"{self.__class__.__name__}.step_finished() called outside of a MLflow run. "
                f"Did you forget to call {self.__class__.__name__}.step_starting() first?"
            )

        step_info = self._running_step_info.get(step.unique_id) or self._get_updated_step_info(step.unique_id)
        if step_info is None:
            raise KeyError(step.unique_id)

        try:
            if step_info.state != StepState.RUNNING:
                raise StepStateError(step, step_info.state)
            step_info.end_time = utc_now_datetime()
            step_info.error = exception_to_string(e)
            terminate_mlflow_run_of_tango_step(
                self.mlflow_client,
                self.experiment_name,
                status="FAILED",
                step_info=step_info,
            )
        finally:
            self.locks[step].release()
            del self.locks[step]
            if step.unique_id in self._running_step_info:
                del self._running_step_info[step.unique_id]

    def register_run(self, targets: Iterable[Step], name: Optional[str] = None) -> Run:
        all_steps = set(targets)
        for step in targets:
            all_steps |= step.recursive_dependencies

        mlflow_run = add_mlflow_run_of_tango_run(
            self.mlflow_client,
            self.experiment_name,
            steps=all_steps,
            run_name=name,
            mlflow_tags=self._mlflow_tags,
        )

        logger.info("Registring run %s with MLflow", name)
        logger.info(
            "View run at: %s/#/experiments/%s/runs/%s",
            self.mlflow_tracking_uri,
            mlflow_run.info.experiment_id,
            mlflow_run.info.run_id,
        )

        run = self.registered_run(mlflow_run.data.tags[MLFLOW_RUN_NAME])

        for step in all_steps:
            self._step_id_to_run_name[step.unique_id] = run.name

        class RunTerminator:
            workspace = self

            def __del__(self) -> None:
                self.workspace.terminate_run(run)

        setattr(run, "_tango_mlflow_run_terminator", RunTerminator())
        atexit.register(self.terminate_run, run)

        return run

    def terminate_run(self, run: Union[str, Run]) -> None:
        run_name = run if isinstance(run, str) else run.name

        mlflow_run = get_mlflow_run_by_tango_run(
            self.mlflow_client,
            self.experiment_name,
            tango_run=run,
            additional_filter_string=" AND ".join(
                (
                    "attributes.status != 'FINISHED'",
                    "attributes.status != 'FAILED'",
                    "attributes.status != 'KILLED'",
                )
            ),
        )
        if mlflow_run is not None:
            status = self._run_status.get(run_name, "FINISHED")
            logger.info("Terminating run %s with status %s", run.name if isinstance(run, Run) else run, status)
            self.mlflow_client.set_terminated(mlflow_run.info.run_id, status=status)

        if run_name in self._run_status:
            self._run_status.pop(run_name)

    def registered_runs(self) -> Dict[str, Run]:
        return {
            mlflow_run.data.tags[MLFLOW_RUN_NAME]: self._get_tango_run_by_mlflow_run(mlflow_run)
            for mlflow_run in get_mlflow_runs(
                self.mlflow_client,
                self.experiment_name,
                run_kind=RunKind.TANGO_RUN,
            )
        }

    def registered_run(self, name: str) -> Run:
        mlflow_run = get_mlflow_run_by_tango_run(self.mlflow_client, self.experiment_name, name)
        if mlflow_run is None:
            raise KeyError(f"Run '{name}' not found in workspace")
        return self._get_tango_run_by_mlflow_run(mlflow_run)

    def _get_tango_run_by_mlflow_run(self, mlflow_run: MLFlowRun) -> Run:
        step_name_to_info: Dict[str, StepInfo] = {}
        with tempfile.TemporaryDirectory() as temp_dir:
            artifact_path = Path(
                mlflow.artifacts.download_artifacts(
                    run_id=mlflow_run.info.run_id,
                    artifact_path="",
                    dst_path=temp_dir,
                )
            )
            steps_json_path = artifact_path / "steps.json"
            with steps_json_path.open("r") as jsonfile:
                for step_name, step_info_dict in json.load(jsonfile)["steps"].items():
                    step_info = StepInfo.from_json_dict(step_info_dict)
                    if step_info.cacheable:
                        updated_step_info = self._get_updated_step_info(step_info.unique_id, step_name=step_name)
                        if updated_step_info is not None:
                            step_info = updated_step_info
                    step_name_to_info[step_name] = step_info

        run_name = mlflow_run.data.tags[MLFLOW_RUN_NAME]
        return Run(
            name=run_name,
            steps=step_name_to_info,
            start_date=datetime.fromtimestamp(mlflow_run.info.start_time / 1000, tz=pytz.utc),
        )

    def _get_updated_step_info(self, step_id: str, step_name: Optional[str] = None) -> Optional[StepInfo]:
        def load_step_info(mlflow_run: MLFlowRun) -> StepInfo:
            with tempfile.TemporaryDirectory() as temp_dir:
                path = mlflow.artifacts.download_artifacts(
                    run_id=mlflow_run.info.run_id,
                    artifact_path="step_info.json",
                    dst_path=temp_dir,
                )
                with open(path, "r") as jsonfile:
                    return StepInfo.from_json_dict(json.load(jsonfile))

        for mlflow_run in get_mlflow_runs(
            self.mlflow_client,
            self.experiment_name,
            run_kind=RunKind.STEP,
            tango_step=step_id,
            additional_filter_string=None if step_name is None else f"tags.step_name = '{step_name}'",
        ):
            step_info = load_step_info(mlflow_run)
            if step_info.start_time is None:
                step_info.start_time = datetime.fromtimestamp(mlflow_run.info.start_time / 1000, tz=pytz.utc)
            if mlflow_run.info.status in ("FINISHED", "FAILED"):
                if step_info.end_time is None:
                    step_info.end_time = datetime.fromtimestamp(mlflow_run.info.end_time / 1000, tz=pytz.utc)
                if mlflow_run.info.status == "FAILED" and step_info.error is None:
                    step_info.error = "Exception"
            return step_info

        def load_step_info_dict(mlflow_run: MLFlowRun) -> Dict[str, Any]:
            with tempfile.TemporaryDirectory() as temp_dir:
                path = mlflow.artifacts.download_artifacts(
                    run_id=mlflow_run.info.run_id,
                    artifact_path="steps.json",
                    dst_path=temp_dir,
                )
                with open(path, "r") as jsonfile:
                    step_info_dict = json.load(jsonfile)
                    assert isinstance(step_info_dict, dict)
                    return step_info_dict

        for mlflow_run in get_mlflow_runs(
            self.mlflow_client,
            self.experiment_name,
            run_kind=RunKind.TANGO_RUN,
            tango_step=step_id,
        ):
            step_info_dict = load_step_info_dict(mlflow_run)
            if step_name is not None:
                step_info_data = step_info_dict["steps"][step_name]
            else:
                step_info_data = next(d for d in step_info_dict["steps"].values() if d["unique_id"] == step_id)
            step_info = StepInfo.from_json_dict(step_info_data)
            return step_info

        return None

    def capture_logs_for_run(self, name: str) -> ContextManager[None]:
        @contextmanager
        def capture_logs() -> Iterator[None]:
            with tempfile.TemporaryDirectory() as temp_dir:
                log_path = Path(temp_dir) / "out.log"
                try:
                    with file_handler(log_path) as handler:
                        yield handler
                finally:
                    mlflow_run = get_mlflow_run_by_tango_run(self.mlflow_client, self.experiment_name, name)
                    if mlflow_run is None:
                        raise RuntimeError(f"Run '{name}' not found in workspace")
                    self.mlflow_client.log_artifact(mlflow_run.info.run_id, str(log_path))

        return capture_logs()

    def _to_params(self) -> Dict[str, Any]:
        return {
            "type": "mlflow",
            "experiment_name": self.experiment_name,
            "tags": self._mlflow_tags,
            "tracking_uri": self.mlflow_client.tracking_uri,
        }
